;changelog:
;05.03.07:	sys,ip-addr, changed start address to $c000 !
;28.02.07: 	call returns reg values
;28.02.07:	rewritten for ca65
;---------------------------------------
;netmon
;
;rr-net monitoring server
;oct.06 hannenz@freenet.de
;based on udpslave by magervalp
;assemble with turbomacropro/style.
;---------------------------------------

packet_page	= $de02
packet_data	= $de04
rxtx_data	= $de08
tx_cmd		= $de0c
tx_len		= $de0e

ptr		= $2d
len		= $57
ip_cksum	= $fb
ip_len		= $fd
packet_len	= $ff

BUF_SIZE	= 256
;---------------------------------------
.code

	.word *+2
;---------------------------------------
;here we go...

;check for ip-adress after SYS
	jsr $79
	cmp #','
	bne init

	
	lda #0
	sta ptr

:	jsr $aefd
	jsr $b79e
	txa
	ldx ptr
	sta cfg_ip,x
	inc ptr
	lda ptr
	cmp #4
	bne :-

init:
;init retro replay, device and memory access
	lda #<mssg
	ldy #>mssg
	jsr $ab1e

	lda #0
	sta ptr

:	ldy ptr
	ldx cfg_ip,y
	lda #0
	jsr $bdcd
	lda #'.'
	jsr $ffd2
	inc ptr
	lda ptr
	cmp #4
	bne :-
	lda #13
	jsr $ffd2

	lda #8     ;default device
	sta dev
	lda #$37   ;access rom/io
	sta access
	sta 1

;detect rr-net

	lda #0
	jsr read_page
	cpx #$0e
	bne notfound
	cpy #$63
	bne notfound

	ldy #$6b
	jsr $f12f   ;print "ok"

;reset rr-net and do some init

	lda #$8a
	ldx #$55
	ldy #0
	jsr write_page
	lda #$82
	ldx #$05
	ldy #$0d
	jsr write_page
	lda #$ac
	ldx cfg_mac
	ldy cfg_mac+1
	jsr write_page
	lda #$ad
	ldx cfg_mac+2
	ldy cfg_mac+3
	jsr write_page
	lda #$ae
	ldx cfg_mac+4
	ldy cfg_mac+5
	jsr write_page
	lda #$89
	ldx #$d3
	ldy #0
	jsr write_page

	jmp main

notfound:
	lda #$ff   ;"device not pr."
	sta $3a
	ldx #5
	jmp $a437

;---------------------------------------
;read a rr-net register reg/2 in .a
;value in .x/.y

read_page:
	asl a
	sta packet_page
	lda #0
	rol a
	sta packet_page+1
	ldx packet_data
	ldy packet_data+1
	rts

;---------------------------------------
;write a rr-net register re/2 in. a
;returns value in .x/.y

write_page:
	asl
	sta packet_page
	lda #0
	rol
	sta packet_page+1
	stx packet_data
	sty packet_data+1
	rts

;---------------------------------------
;flush buffer and fall back to main
;loop

purge:	
	jsr read_word
	jmp purge

;---------------------------------------
;read two bytes from rr-net into .x/.y

read_word:
	lda len+1
	bmi @abort
	bne :+
	lda len
	beq @abort
:	lda len
	sec
	sbc #2
	sta len
	bcs :+
	dec len+1
:	lda rxtx_data
	ldx rxtx_data+1
	rts

@abort:
	ldx stack
	txs

;---------------------------------------
;main loop: wait for incoming packets
;and react upon
;---------------------------------------

main:
	tsx
	stx stack

	;wait for incoming packet

	lda #$24
	sta packet_page
	lda #$01
	sta packet_page+1

:	lda packet_data+1
	and #$0d
	bne :+

	;user break...?! (stop)

	lda $91
	cmp #$7f
	bne :-
	jmp $a84b    ;basic break

	;ignore status
	;important! read hi byte first for status and length

	;-- rr-net chip info --

:	lda rxtx_data+1
	lda rxtx_data

	;get packet length
	lda rxtx_data+1
	sta len+1
	lda rxtx_data
	sta len

	;-- mac layer (ethernet) --

	;skip dest mac
	ldy #3
	jsr readmultiword

	;who is calling...?!
	jsr readpcmac

	;type-field
	jsr read_word
	cmp #8
	bne @purge
	cpx #6
	beq arp
	cpx #0
	beq ip

@purge:
	jmp purge

;---------------------------------------
;process an arp packet
arp:
	;discard 1st 6 bytes
	ldy #4
	jsr readmultiword

	;arp command: is it a request...?!
	cpx #1
	bne @purge

	jsr readpcmac
	jsr readpcip

	;discard target mac address (will be 'broadcast' anyway..)
	ldy #4
	jsr readmultiword

	;read target ip address & ignore if not our own one

	cmp cfg_ip
	bne @purge
	cpx cfg_ip+1
	bne @purge
	jsr read_word
	cmp cfg_ip+2
	bne @purge
	cpx cfg_ip+3
	bne @purge

	ldy #6         ;type: arp
	ldx #42        ;length
	jsr startreply

	;send header
	ldy #0
:	lda arpreply,y
	iny
	ldx arpreply,y
	jsr send_word
	iny
	cpy #8
	bne :-

	jsr sendmymac
	jsr sendmyip
	jsr sendpcmac
	jsr sendpcip
	jmp main

@purge:	
	jmp purge


;---------------------------------------
.code
;process ip v4 header
ip:
	;discard version,ihl,tos
	ldy #2
	jsr readmultiword

	;read length
	stx ip_len
	sta ip_len+1

	;substract ip header length --> packet length
	lda ip_len
	ldx ip_len+1
	sec
	sbc #20
	bcs :+
	dex
:	sta packet_len

	;restrict to 256 bytes
	cpx #0
	bne @purge

	jsr read_word
	sta ip_id
	stx ip_id+1

	;discard flags,frag,ttl
	ldy #2
	jsr readmultiword

	;read protocol
	txa
	pha

	;read checksum
	jsr read_word
	sta ip_cksum
	stx ip_cksum+1

	jsr readpcip

	;discard destination ip
	;maybe include check here if the packet is for us...???
	ldy #2
	jsr readmultiword

	pla     ;protocol
	cmp #17
	beq udp

@purge:
	jmp purge

;---------------------------------------
;process udp header...
udp:
	;read ports from header

	jsr read_word
	sta udpsrcport
	stx udpsrcport+1
	jsr read_word
	sta udpdstport
	stx udpdstport+1

	;skip length  & checksum

	ldy #2
	jsr readmultiword

	;adjust packet length and
	;make even

	lda packet_len
	sec
	sbc #8
	sta packet_len
	lsr
	bcc :+
	inc packet_len

	;read data and store in
	;data buffer

:	ldy #0
:	jsr read_word
	sta data,y
	txa
	sta data+1,y
	iny
	iny
	cpy packet_len
	bne :-

@flush:
	lda len+1
	bmi :++
	bne :+
	lda len
	beq :++

:	jsr read_word
	jmp @flush

	;check for port 3172
:	lda udpdstport
	ldx udpdstport+1
	cmp #$0c
	bne @tomain
	cpx #$64
	bne @tomain

;now we have the data in data buf. and we know that the data is for us!
;so - let's rock!

	;act upon first data byte:
	lda data
	asl
	tax
	lda cmdtab+1,x
	pha
	lda cmdtab,x
	pha
	rts

@tomain:
	jmp main

cr:	lda #13
	.byte $2c
dot:	lda #'.'
	.byte $2c
colon:	lda #':'
	.byte $2c
space:	lda #' '
	jmp $ffd2

decnum:	lda #0
	jmp $bdcd
;---------------------------------------
;send string to current dev, read
;command channel and send output back
;1 = len
;2... string

setlfs	= $fe00
setnam	= $fdf9
open	= $f34a
close	= $f291
chkin	= $f20e
ckout	= $f250
clrch	= $f333

bsout	= $f1ca
chrin	= $f157


diskcommand:
	lda #$37
	sta 1
	jsr $f32f  	;clall

	lda #15    	;command channel
	tay
	ldx dev
	jsr setlfs
	lda #0     	;no filename
	jsr setnam
	jsr open	;open

	lda data+1
	beq @skip

	ldx #15
	jsr ckout

	ldy #0
	sty temp
:	ldy temp
	lda data+2,y
	jsr bsout
	inc temp
	lda temp
	cmp data+1
	bne :-

	jsr clrch

@skip:
	lda #0
	sta datalen
	ldx #15
	jsr chkin

:	lda $90
	bne :+
	jsr chrin
	jsr putbyte
	jmp :-

:	lda #0     ;terminate with 0
	jsr putbyte
	jsr clrch
	lda #15
	jsr close
	jsr senddata
	jmp main

;---------------------------------------
;set access type (ram, rom, reu)

setram:
	lda data+1
	sta access
	jmp main

;---------------------------------------

fill:
	lda data+1
	ldx data+2
	ldy data+3
	stx ptr
	sty ptr+1
	ldy #0
:	sta (ptr),y
	inc ptr
	bne :+
	inc ptr+1
:	ldx ptr
	cpx data+4
	bne :--
	ldx ptr+1
	cpx data+5
	bne :--
	jmp main
;---------------------------------------
poke:
	ldx data+2
	ldy data+3
	stx ptr
	sty ptr+1
	ldy #0
:	lda data+4,y
	sta (ptr),y
	iny
	cpy data+1
	bne :-
	jmp main
;---------------------------------------
telldev:
	lda dev
	sta data
	lda #2
	sta datalen
	jsr senddata
	jmp main

;---------------------------------------
setdev:
	lda data+1
	sta dev
	jmp main
;---------------------------------------
;open write file and reply ok or fail
;1:  fnlen
;2.. string
write:
	ldx dev
	lda #1
	tay
	jsr $ffba
	lda data+1
	ldx #<(data+2)
	ldy #>(data+2)
	jsr $ffbd
	jsr $ffc0
	lda $90
	bne :+

	jsr sendok
	jmp main

:	jsr sendfail
	jmp main
;---------------------------------------
;write data to writefile
;1:  number of bytes to write        
;2.. data
;reply ok or fail

xwrite:
	ldx #1
	jsr $ffc9
	bcc :+

	jsr sendfail
	jmp main

:	ldx #0
:	lda data+2,x
	jsr $ffd2
	inx
	cpx data+1
	bne :-

	jsr $ffcc

	jsr sendok
	jmp main
;---------------------------------------
;close write file, reply ok or fail
xdone:
	jsr $ffcc
	lda #1
	jsr $ffc3
	jmp main
;---------------------------------------
;reply ok or fail
sendfail:
	jsr $ffcc   ;close file
	lda #1
	jsr $ffc3
	lda #$ff    ;and send "fail"
	.byte $2c
sendok:	
	lda #0      ;send "ok"
	sta data
	lda #2
	sta datalen
	jmp senddata

;---------------------------------------
;read from current device and send data
read:
	lda #0
	sta datalen

	;open file

	lda #8
	tay
	ldx dev
	jsr $ffba
	lda data+1
	ldx #<(data+2)
	ldy #>(data+2)
	jsr $ffbd
	jsr $ffc0
	ldx #8
	jsr $ffc6

	;send data

:	lda $90
	bne :+
	jsr $ffcf
	jsr putbyte
	jmp :-

	;close file

:	jsr $ffcc
	lda #8
	jsr $ffc3
	jsr senddata
	jmp main
;---------------------------------------
;send directory of current dev
dir:
	lda #0
	sta datalen

	lda data+1	;length of filename
	sta $b7
	lda #<(data+2)
	sta $bb
	lda #>(data+2)
	sta $bc

	lda dev
	sta $ba
	lda #$60	;sec addr.
	sta $b9
	jsr $f3d5
	lda $ba
	jsr $ffb4  ;dev: talk
	lda #$b9
	jsr $ff96

	lda #0
	sta $90
	ldy #3

@loop:	sty ptr
	jsr $ffa5
	sta ptr+1
	ldy $90
	bne @close
	jsr $ffa5
	ldy $90
	bne @close
	ldy ptr
	dey
	bne @loop

	ldy ptr+1
	jsr $b391
	jsr $bddd

	ldx #0
	stx temp
:	ldx temp
	lda $0100,x
	beq :+
	jsr putbyte
	inc temp
	jmp :-

:	lda #$20
	jsr putbyte

:	jsr $ffa5
	ldx $90
	bne @close
	tax
	beq @cr
	jsr putbyte
	jmp :-

@cr:	lda #13
	jsr putbyte
	lda #10
	jsr putbyte
	ldy #2
	bne @loop

@close:
	jsr $f642
	jsr senddata
	jmp main
;---------------------------------------
;put a byte to output buffer and
;send over rr-net if buffer is full
;(max. 128 bytes)

putbyte:
	ldx datalen
	sta data,x
	inc datalen
	bpl :+
	jsr senddata
	lda #0
	sta datalen
:	rts

;---------------------------------------
;transfer memory peek to rr-net
;1:   length
;2-5: address (might be reu)
;6:   data
get:
	lda data+1
	sta datalen

	dec $d020

	bit access
	bmi @getreu

	sei
	lda 1
	pha
	lda access
	sta $01

	lda data+3
	sta ptr
	lda data+4
	sta ptr+1

	ldy #0
:	lda (ptr),y
	sta data,y
	iny
	cpy datalen
	bne :-

	pla
	sta $01
	cli

@out:
	jsr senddata
	inc $d020
	jmp main

	;get data from reu

@getreu:
	lda #<data
	ldx #>data
	jsr setupreu
	lda #253      ;fetch
	sta $df01
	bne @out

;---------------------------------------
;setup reu registers for data at .a/.x
;and reu address at data offset 3-5 and
;length in data offset 1

setupreu:
	sta $df02
	stx $df03
	lda data+3
	sta $df04
	lda data+4
	sta $df05
	lda data+5
	sta $df06
	lda data+1
	sta $df07
	lda #0
	sta $df08
	rts

;---------------------------------------
;send ip-header,udp-header and data
;--> send one frame
;number of data bytes to send must be in
;datalen

senddata:

	;make length even
	lda datalen
	lsr a
	bcc :+
	inc datalen

	;send headers

:	lda datalen
	jsr send_ip_udp

	;send data

	ldy #0
:	lda data,y
	ldx data+1,y
	jsr send_word
	iny
	iny
	cpy datalen
	bne :-
	rts
;---------------------------------------
;read data and store in ram
;1:   len
;3-5: address (might be reu!)
;6    data
put:
	dec $d020
	bit access
	bpl @putram    ;to c64 ram

	;put to reu

	lda #<(data+6)
	ldx #>(data+6)
	jsr setupreu
	lda #252      ;stash
	sta $df01
	jmp @out

	;load to ram

@putram:
	lda data+3
	sta ptr
	lda data+4
	sta ptr+1
	ldy #0
:	lda data+6,y
	sta (ptr),y
	iny
	cpy data+1
	bne :-

@out:
	jsr sendok
	inc $d020
	jmp main
;---------------------------------------
basic:
	ldx stack
	txs
	ldx #$80
	jmp ($0300)

;---------------------------------------
initialize:
	jmp init

;---------------------------------------
status:
	lda access
	sta data
	lda dev
	sta data+1
	lda #2
	sta datalen
	jsr senddata
	jmp main
;---------------------------------------
;perfrom "run"
run:
	lda #$37
	sta 1
	lda $ae
	sta $2d
	lda $af
	sta $2e
	jsr $a533
	jsr $a659
	jsr $a68e
	ldx #$f6
	txs
	jmp $a7ae

;---------------------------------------
;call:
;1-2: address
;3:   akku
;4:   x-reg
;5:   yreg
;6:   status reg

call:
	lda #>(aftercall-1) ;return to main
	pha            ;after call...
	lda #<(aftercall-1)
	pha
	lda data+6   ;sr
	pha
	lda data+3   ;.a
	ldx data+4   ;.x
	ldy data+5   ;.y
	plp          ;set sr
jump:
	jmp (data+1)

aftercall:
	php
	sta data
	stx data+1
	sty data+2
	pla
	sta data+3
	lda #4
	sta datalen
	jsr senddata
	jmp main



;---------------------------------------
;makes ip & udp header and sends them
;after this you can start transfering
;data
;---------------------------------------
send_ip_udp:

	; 14 : length of mac header
	;+20 : length of ip  header
	;+8  : length of udp header
	;+.a : length of udp data
	;=     length of whole frame
	;      to transfer!

	clc
	adc #28
	pha    ;remember ip length
	clc
	adc #14
	tax     ;len
	ldy #0  ;ip
	jsr startreply

	lda #0
	sta ip_cksum
	sta ip_cksum+1
	sta ip_id
	sta ip_id+1

	; -- calc ip checksum --

	lda #$45
	ldx #0
	jsr cksum_add

	pla            ;ip length
	sta ip_len
	tax
	lda #0
	adc #0
	sta ip_len+1
	jsr cksum_add

	lda #$80       ;ttl
	ldx #$11       ;protocol
	jsr cksum_add

	lda cfg_ip     ;source ip
	ldx cfg_ip+1
	jsr cksum_add
	lda cfg_ip+2
	ldx cfg_ip+3
	jsr cksum_add
	lda pc_ip      ;dest ip
	ldx pc_ip+1
	jsr cksum_add
	lda pc_ip+2
	ldx pc_ip+3
	jsr cksum_add

	lda ip_cksum   ;invert cksum
	eor #$ff
	sta ip_cksum
	lda ip_cksum+1
	eor #$ff
	sta ip_cksum+1

	; -- ip v4 --

	lda #$45      ;vers./header len
	ldx #$00      ;type of service
	jsr send_word
	lda ip_len+1  ;total length
	ldx ip_len
	jsr send_word
	lda ip_id     ;identification
	ldx ip_id+1
	jsr send_word
	lda #0        ;flags
	tax
	jsr send_word
	lda #$80      ;ttl
	ldx #$11      ;protococl (udp)
	jsr send_word
	lda ip_cksum  ;header checksum
	ldx ip_cksum+1
	jsr send_word
	jsr sendmyip  ;source ip
	jsr sendpcip  ;dest ip

	;--- udp -----

	lda udpdstport   ;source port
	ldx udpdstport+1
	jsr send_word
	lda udpsrcport   ;dest port
	ldx udpsrcport+1
	jsr send_word

	lda ip_len       ;udp len
	sec
	sbc #20
	tax
	lda ip_len+1
	sbc #0
	jsr send_word
	lda #0         ;checksum empty
	tax
;---------------------------------------
send_word:
	sta rxtx_data
	stx rxtx_data+1
	rts

;---------------------------------------
;tells chip to transfer .a bytes,
;sends mac header and protocol in .y
;---------------------------------------
startreply:
	lda #$c9
	sta tx_cmd
	lda #0
	sta tx_cmd+1

	stx tx_len
	sta tx_len+1

	lda #<$0138
	sta packet_page
	lda #>$0138
	sta packet_page+1

:	lda packet_data+1
	ldx packet_data
	lsr a
	bcc :-

	jsr sendpcmac
	jsr sendmymac

	lda #8
	sta rxtx_data
	sty rxtx_data+1
	tya
	rts
;---------------------------------------
sendpcmac:
	ldx #0
:	lda pc_mac,x
	sta rxtx_data
	inx
	lda pc_mac,x
	sta rxtx_data+1
	inx
	cpx #6
	bne :-
	rts
;---------------------------------------
sendmymac:
	ldx #0
:	lda cfg_mac,x
	sta rxtx_data
	inx
	lda cfg_mac,x
	sta rxtx_data+1
	inx
	cpx #6
	bne :-
	rts
;---------------------------------------
sendpcip:
	ldx #0
:	lda pc_ip,x
	sta rxtx_data
	inx
	lda pc_ip,x
	sta rxtx_data+1
	inx
	cpx #4
	bne :-
	rts
;---------------------------------------
sendmyip:
	ldx #0
:       lda cfg_ip,x
	sta rxtx_data
	inx
	lda cfg_ip,x
	sta rxtx_data+1
	inx
	cpx #4
	bne :-
	rts
;---------------------------------------
readmultiword:
	jsr read_word
	dey
	bne readmultiword
	rts

;---------------------------------------
readpcmac:
	ldy #0
:	jsr read_word
	sta pc_mac,y
	txa
	sta pc_mac+1,y
	iny
	iny
	cpy #6
	bne :-
	rts
;---------------------------------------
readpcip:
	ldy #0
:	jsr read_word
	sta pc_ip,y
	txa
	sta pc_ip+1,y
	iny
	iny
	cpy #4
	bne :-
	rts
;---------------------------------------
cksum_add:
	pha
	txa
	clc
	adc ip_cksum+1
	sta ip_cksum+1
	pla
	adc ip_cksum
	sta ip_cksum
	bcc :+
	inc ip_cksum+1
:	rts
;---------------------------------------
;just for debugging...
prhex:
	pha
	lsr
	lsr
	lsr
	lsr
	jsr :+
	pla
	and #$0f
:	clc
	adc #$30
	cmp #$3a
	bcc :+
	adc #6
:	jmp $ffd2
;---------------------------------------
;.rodata
;---------------------------------------
.rodata
mssg:		.byte "netserv v0.3",13,0
arpreply:	.byte 0,1,8,0,6,4,0,2  ;header constants
cfg_mac:	.byte $00,$80,$10,$b3,$eb,$b7
cfg_ip:		.byte 192,168,0,64
dollar:		.byte '$'
cmdtab:	
		.addr initialize-1
		.addr put-1
		.addr get-1
		.addr read-1
		.addr write-1
		.addr xwrite-1
		.addr xdone-1
		.addr run-1
		.addr jump-1
		.addr call-1
		.addr basic-1
		.addr dir-1
		.addr diskcommand-1
		.addr setram-1
		.addr poke-1
		.addr telldev-1
		.addr setdev-1
		.addr fill-1
		.addr status-1
;---------------------------------------
;.bss
;---------------------------------------
.bss
stack:		.res 1
datalen:	.res 1
temp:		.res 1
dev:		.res 1
access:		.res 1
pc_mac:		.res 6
pc_ip:		.res 4
ip_id:		.res 2 
icmp_cksum:	.res 2
udpdstport:	.res 2
udpsrcport:	.res 2
data:		.res BUF_SIZE
;---------------------------------------
